@@ -7,8 +7,8 @@ from scipy.special import gammaln
 from svolfit.models.svol_model import svol_model
 from svolfit.models.model_utils import meanvariance,logsumexp
 
-from svolfit.models.Heston_utils import stepcalc_tree,condassetprob_calc_limit,condassetprob_calc_exact,condassetprob_calc_limit_X2
-from svolfit.models.Heston_utils import Heston_lncondassetprob,Heston_griddefs,Heston_pathprob
+from svolfit.models.Heston_utils import stepcalc_tree,condassetprob_calc_exact,condassetprob_calc_limit_X2
+from svolfit.models.Heston_utils import Heston_lncondassetprob,Heston_griddefs,Heston_pathprob,Heston_UsefulGrid,Heston_standardreturn
 
 #---------------------------------------------
 
@@ -28,16 +28,19 @@ class Heston_tree(svol_model):
         xi=1.0
         u0=1
         
-        self.lamb=0.4
+        self.gridfactor=0.4
 
         self.workingpars_names=['mu','sigma','rho','alpha','xi','u0']
         self.workingpars=np.array([mu,sigma,rho,alpha,xi,u0])
+        self.workingpars_sim=np.array([mu,sigma,rho,alpha,xi,u0])
         self.workingpars_diffs=[0.0001,0.0001,0.0005,0.001,0.0001,0.0001]
 
         alpha_min=np.minimum(2.0/((self.Nobs-1)/252.0),1.0)
 #                 [hmu, hsigma, rho, alpha, xi,u0]
         self.workingpars_bounds=[(-0.5,0.5), (0.05, 0.5), (-0.9,0.9), (alpha_min, 20.0), (0.1, 3.5),(0.1,3.0)]
 
+        self.workingpars_optflag=[True for x in self.workingpars]
+
         if 'init' in self.options:
             self.initpars_reporting(self.options['init'])
 
@@ -97,22 +100,30 @@ class Heston_tree(svol_model):
         xi=eta/sigma
         u0=v0/theta
         
-        self.workingpars[0]=mu
-        self.workingpars[1]=sigma
-        self.workingpars[2]=rho
-        self.workingpars[3]=alpha
-        self.workingpars[4]=xi
-        self.workingpars[5]=u0
-
+        if( pardict['type']=='init' ):
+            self.workingpars[0]=mu
+            self.workingpars[1]=sigma
+            self.workingpars[2]=rho
+            self.workingpars[3]=alpha
+            self.workingpars[4]=xi
+            self.workingpars[5]=u0
+        else:
+            self.workingpars_sim[0]=mu
+            self.workingpars_sim[1]=sigma
+            self.workingpars_sim[2]=rho
+            self.workingpars_sim[3]=alpha
+            self.workingpars_sim[4]=xi
+            self.workingpars_sim[5]=u0
+    
         return
 
     def get_structure(self):
         assetname='asset'
         variancename='variance'
 
-        sigma=self.workingpars[1]
-        rho=self.workingpars[2]
-        u0=self.workingpars[5]
+        sigma=self.workingpars_sim[1]
+        rho=self.workingpars_sim[2]
+        u0=self.workingpars_sim[5]
 
         corrmatrix=np.array([[1.0,rho],[rho,1.0]])
 #TODO: best choice based on pars?
@@ -125,11 +136,11 @@ class Heston_tree(svol_model):
 
     def sim_step(self,asset,variance,Zs):
 
-        mu=self.workingpars[0] 
-        sigma=self.workingpars[1]
-#        rho=self.workingpars[2]
-        alpha=self.workingpars[3] 
-        xi=self.workingpars[4]
+        mu=self.workingpars_sim[0] 
+        sigma=self.workingpars_sim[1]
+#        rho=self.workingpars_sim[2]
+        alpha=self.workingpars_sim[3] 
+        xi=self.workingpars_sim[4]
 
         theta=sigma*sigma
         eta=xi*sigma
@@ -180,97 +191,123 @@ class Heston_tree(svol_model):
         ret['rep_rho']=rho
         ret['rep_alpha']=alpha
         ret['rep_eta']=eta
-        ret['misc_q']=q
 #        ret['u0']=u0
-#        ret['uT']=uT
         ret['rep_v0']=v0
-        ret['rep_vT']=vT
+
+
+        sim_mu=self.workingpars_sim[0] 
+        sim_sigma=self.workingpars_sim[1]
+        sim_rho=self.workingpars_sim[2]
+        sim_alpha=self.workingpars_sim[3] 
+        sim_xi=self.workingpars_sim[4]
+        sim_u0=self.workingpars_sim[5]
+
+        sim_theta=sim_sigma*sim_sigma
+        sim_eta=sim_xi*sim_sigma
+        sim_v0=sim_sigma*sim_sigma*sim_u0
+
+        ret['sim_wrk_mu']=sim_mu
+        ret['sim_wrk_sigma']=sim_sigma
+        ret['sim_wrk_rho']=sim_rho
+        ret['sim_wrk_alpha']=sim_alpha
+        ret['sim_wrk_xi']=sim_xi
+        ret['sim_wrk_u0']=sim_u0
+
+        ret['sim_rep_mu']=sim_mu
+        ret['sim_rep_theta']=sim_theta
+        ret['sim_rep_rho']=sim_rho
+        ret['sim_rep_alpha']=sim_alpha
+        ret['sim_rep_eta']=sim_eta
+        ret['sim_rep_v0']=sim_v0
+
+        ret['misc_q']=q
+#        ret['uT']=uT
+        ret['misc_vT']=vT
         ret['ts_vpath']=vpath
         ret['ts_upath']=self.upath
+
+        Nret=self.Nobs-1
+        self.standardreturns=np.zeros(Nret)
+        Heston_standardreturn(self.yasset,self.upath,self.dt,rho,sigma,mu,alpha,xi,self.standardreturns)
+        ret['ts_standardreturns']=self.standardreturns
         
         return ret
 
-    def workingpars_update(self,workingpars):
-        super().workingpars_update(workingpars)
+    def update(self):
+        super().update()
         
-        if( self.current==False):
-# update all grid quantities to be cached:
-#TODO: struct-ify?
-            mu=self.workingpars[0] 
-            sigma=self.workingpars[1]
-            rho=self.workingpars[2]
-            alpha=self.workingpars[3] 
-            xi=self.workingpars[4]
-            u0=self.workingpars[5]
-            q=2.0*alpha/(xi*xi)
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+        u0=self.workingpars[5]
+        q=2.0*alpha/(xi*xi)
 
-            dt=self.dt
-            
-            if( alpha*dt < 1.0e-6 ):
-                self.gamm=0.5*alpha*dt*(1+0.5*alpha*dt)
-                self.Vb=(xi*xi)*np.exp(-alpha*dt)*dt*(1-0.5*alpha*dt)
-            else:
-                self.gamm=np.expm1(alpha*dt)/2.0
-                self.Vb=-xi*xi*np.expm1(-alpha*dt)*np.exp(-alpha*dt)/alpha
-            if( np.isnan(self.gamm) or np.isnan(self.Vb) ):
-                print('NAN ERROR: gamm,Vb')
-                print(alpha,q,xi,self.gamm,self.Vb)
-            
-            self.delta=self.Vb/(2.0*self.lamb)
-            
-            t=np.linspace(1,self.Nobs,self.Nobs)*dt
-            uj=-np.expm1(-alpha*t)+np.exp(-alpha*t)*u0
+        dt=self.dt
         
-            deltajrat=(np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-0.5)
-            self.nj=np.array( np.ceil( np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-np.sqrt(0.25+2.0*self.gamm/self.delta ) ) , int)
+        if( alpha*dt < 1.0e-6 ):
+            self.gamm=0.5*alpha*dt*(1+0.5*alpha*dt)
+            self.Vb=(xi*xi)*np.exp(-alpha*dt)*dt*(1-0.5*alpha*dt)
+        else:
+            self.gamm=np.expm1(alpha*dt)/2.0
+            self.Vb=-xi*xi*np.expm1(-alpha*dt)*np.exp(-alpha*dt)/alpha
+        if( np.isnan(self.gamm) or np.isnan(self.Vb) ):
+            print('NAN ERROR: gamm,Vb')
+            print(alpha,q,xi,self.gamm,self.Vb)
         
-            self.tu0=uj+self.delta*(-self.nj*deltajrat+self.nj*(self.nj-1)/2.0)
-            self.tdelta0=self.delta*(deltajrat-self.nj)
-            
-            maxindex = self.Nobs+np.max(self.nj)
+        self.delta=self.Vb/(2.0*self.gridfactor)
+        
+        t=np.linspace(1,self.Nobs,self.Nobs)*dt
+        uj=-np.expm1(-alpha*t)+np.exp(-alpha*t)*u0
+    
+        deltajrat=(np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-0.5)
+        self.nj=np.array( np.ceil( np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-np.sqrt(0.25+2.0*self.gamm/self.delta ) ) , int)
+    
+        self.tu0=uj+self.delta*(-self.nj*deltajrat+self.nj*(self.nj-1)/2.0)
+        self.tdelta0=self.delta*(deltajrat-self.nj)
+        
+        maxindex = self.Nobs+np.max(self.nj)
 
 # maxindex can depend on parameters, so have to reallocate all the time...
-            self.ni_s=np.zeros(maxindex+1,dtype=int)
-            self.space=np.zeros((7,maxindex+1))
-            self.pi_s=np.zeros((12,maxindex+1))
-            self.u_s=np.zeros((3,maxindex+1))
-            self.p_s=np.zeros((2,maxindex+1))
-            
-            CondProbCalc='Limit'
-            if( CondProbCalc=='Limit' ):
-                coeff_dt=(mu-alpha*rho*sigma/xi)*dt
-                coeff_iu = rho*alpha*sigma/xi -sigma*sigma/2.0
-                coeff_du = rho*sigma/xi
-                self.condprob_calc=lambda yasset,u_prev,u_this: condassetprob_calc_limit(yasset,u_prev,u_this,dt,rho,sigma,coeff_dt,coeff_iu,coeff_du)
-            elif( CondProbCalc=='Exact' ):
-        #TODO: If you really want to use this then it needs to be cleaned up and reviewed!!
-                coeff_dt=(mu-alpha*rho*sigma/xi)*dt
-                coeff_iu = rho*alpha*sigma/xi -sigma*sigma/2.0
-                coeff_du = rho*sigma/xi
+        self.ni_s=np.zeros(maxindex+1,dtype=int)
+        self.space=np.zeros((7,maxindex+1))
+        self.pi_s=np.zeros((12,maxindex+1))
+        self.u_s=np.zeros((3,maxindex+1))
+        self.p_s=np.zeros((2,maxindex+1))
         
-                epsilon=alpha*dt/2.0
-                MI_f1=1.0-2.0*epsilon*epsilon/15.0+2.0*epsilon*epsilon*epsilon*epsilon/105.0
-                MI_f2=1.0-epsilon*epsilon/15.0+2.0*epsilon*epsilon*epsilon*epsilon/315.0
-                MI_f3=1.0-7.0*epsilon*epsilon/30.0+31.0*epsilon*epsilon*epsilon*epsilon/840.0
-                if( epsilon > 1.0e-4 ):
-                    MI_f1=3.0*(np.cosh(epsilon)/(epsilon*np.sinh(epsilon))-1.0/(np.sinh(epsilon)**2))/2.0
-                    MI_f2=3.0*(epsilon*np.cosh(epsilon)/np.sinh(epsilon)-1.0)/(epsilon**2)
-                    MI_f3=3.0*(epsilon*np.cosh(epsilon)/np.sinh(epsilon)-1.0)/(epsilon*np.sinh(epsilon))
+        CondProbCalc='Limit'
+        if( CondProbCalc=='Limit' ):
+            self.lncondprob_calc=lambda yasset,u_prev,u_this,lncp: Heston_lncondassetprob(yasset,u_prev,u_this,self.dt,rho,sigma,mu,alpha,xi,lncp)
+        elif( CondProbCalc=='Exact' ):
+    #TODO: If you really want to use this then it needs to be cleaned up and reviewed!!
+            # coeff_dt=(mu-alpha*rho*sigma/xi)*dt
+            # coeff_iu = rho*alpha*sigma/xi -sigma*sigma/2.0
+            # coeff_du = rho*sigma/xi
+    
+            # epsilon=alpha*dt/2.0
+            # MI_f1=1.0-2.0*epsilon*epsilon/15.0+2.0*epsilon*epsilon*epsilon*epsilon/105.0
+            # MI_f2=1.0-epsilon*epsilon/15.0+2.0*epsilon*epsilon*epsilon*epsilon/315.0
+            # MI_f3=1.0-7.0*epsilon*epsilon/30.0+31.0*epsilon*epsilon*epsilon*epsilon/840.0
+            # if( epsilon > 1.0e-4 ):
+            #     MI_f1=3.0*(np.cosh(epsilon)/(epsilon*np.sinh(epsilon))-1.0/(np.sinh(epsilon)**2))/2.0
+            #     MI_f2=3.0*(epsilon*np.cosh(epsilon)/np.sinh(epsilon)-1.0)/(epsilon**2)
+            #     MI_f3=3.0*(epsilon*np.cosh(epsilon)/np.sinh(epsilon)-1.0)/(epsilon*np.sinh(epsilon))
+    
+            # self.condprob_calc=lambda yasset,u_prev,u_this: condassetprob_calc_exact(yasset,u_prev,u_this,dt,alpha,rho,sigma,xi,q,MI_f1,MI_f2,MI_f3,coeff_dt,coeff_iu,coeff_du)
+    
+            pass
+        else:
+            print('Unknown Calc -- Handle')                
         
-                self.condprob_calc=lambda yasset,u_prev,u_this: condassetprob_calc_exact(yasset,u_prev,u_this,dt,alpha,rho,sigma,xi,q,MI_f1,MI_f2,MI_f3,coeff_dt,coeff_iu,coeff_du)
-        
-                pass
-            else:
-                print('Unknown Calc -- Handle')                
-            
 
         return
 
     def calculate(self,OptRun=True):
         
-#        mu=self.workingpars[0] 
-#        sigma=self.workingpars[1]
-#        rho=self.workingpars[2]
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
         alpha=self.workingpars[3] 
         xi=self.workingpars[4]
         u0=self.workingpars[5]
@@ -343,11 +380,16 @@ class Heston_tree(svol_model):
                 u_this = u_s[0,0:n_top2]
                 p_this = p_s[0,0:n_top2]
                 iseven=True
-    
-            condprob_mid[:]=self.condprob_calc(yasset[cc],u_prev,u_this[ni-n_min])
-            condprob_dn[:]=self.condprob_calc(yasset[cc],u_prev,u_this[ni-n_min-1])
-            condprob_up[:]=self.condprob_calc(yasset[cc],u_prev,u_this[ni-n_min+1])
-    
+ 
+# recall the s=reshape makes the array Nobs=1 x Ngrid so that the same function call will work
+# I believe without making a copy of memory on the array...
+            self.lncondprob_calc(yasset[[cc]],u_prev,u_this[ni-n_min]  ,np.reshape(condprob_mid,(1,n_top)))
+            self.lncondprob_calc(yasset[[cc]],u_prev,u_this[ni-n_min-1],np.reshape(condprob_dn ,(1,n_top)))
+            self.lncondprob_calc(yasset[[cc]],u_prev,u_this[ni-n_min+1],np.reshape(condprob_up ,(1,n_top)))
+            condprob_mid[:]=np.exp(condprob_mid)
+            condprob_dn[:]=np.exp(condprob_dn)
+            condprob_up[:]=np.exp(condprob_up)
+
             p_this[:]=np.zeros(n_top2)
             if( OptRun==False ):
                 vp=np.zeros(n_top2)
@@ -458,6 +500,14 @@ class Heston_tree(svol_model):
         self.calculate(False)
         return 
 
+    def status(self):
+        if self.current != True:
+            return 'Failure','Incorrect Model State.'
+
+        status = 'Success'
+        message = 'No issues.'        
+        return (status,message)
+
 #---------------------------------------------
 
 class Heston_treeX2(svol_model):
@@ -476,17 +526,20 @@ class Heston_treeX2(svol_model):
         xi=1.0
         u0=1
         
-        self.lamb=0.4
+        self.gridfactor=0.4
         self.dt2 = self.dt/2.0
 
         self.workingpars_names=['mu','sigma','rho','alpha','xi','u0']
         self.workingpars=np.array([mu,sigma,rho,alpha,xi,u0])
+        self.workingpars_sim=np.array([mu,sigma,rho,alpha,xi,u0])
         self.workingpars_diffs=[0.0001,0.0001,0.0005,0.001,0.0001,0.0001]
 
         alpha_min=np.minimum(2.0/((self.Nobs-1)/252.0),1.0)
 #                 [hmu, hsigma, rho, alpha, xi,u0]
         self.workingpars_bounds=[(-0.5,0.5), (0.05, 0.5), (-0.9,0.9), (alpha_min, 20.0), (0.1, 4.5),(0.1,3.0)]
 
+        self.workingpars_optflag=[True for x in self.workingpars]
+
         if 'init' in self.options:
             self.initpars_reporting(self.options['init'])
 
@@ -546,22 +599,30 @@ class Heston_treeX2(svol_model):
         xi=eta/sigma
         u0=v0/theta
         
-        self.workingpars[0]=mu
-        self.workingpars[1]=sigma
-        self.workingpars[2]=rho
-        self.workingpars[3]=alpha
-        self.workingpars[4]=xi
-        self.workingpars[5]=u0
-
+        if( pardict['type']=='init' ):
+            self.workingpars[0]=mu
+            self.workingpars[1]=sigma
+            self.workingpars[2]=rho
+            self.workingpars[3]=alpha
+            self.workingpars[4]=xi
+            self.workingpars[5]=u0
+        else:
+            self.workingpars_sim[0]=mu
+            self.workingpars_sim[1]=sigma
+            self.workingpars_sim[2]=rho
+            self.workingpars_sim[3]=alpha
+            self.workingpars_sim[4]=xi
+            self.workingpars_sim[5]=u0
+            
         return
 
     def get_structure(self):
         assetname='asset'
         variancename='variance'
 
-        sigma=self.workingpars[1]
-        rho=self.workingpars[2]
-        u0=self.workingpars[5]
+        sigma=self.workingpars_sim[1]
+        rho=self.workingpars_sim[2]
+        u0=self.workingpars_sim[5]
 
         corrmatrix=np.array([[1.0,rho],[rho,1.0]])
 #TODO: best choice based on pars?
@@ -574,11 +635,11 @@ class Heston_treeX2(svol_model):
 
     def sim_step(self,asset,variance,Zs):
 
-        mu=self.workingpars[0] 
-        sigma=self.workingpars[1]
-#        rho=self.workingpars[2]
-        alpha=self.workingpars[3] 
-        xi=self.workingpars[4]
+        mu=self.workingpars_sim[0] 
+        sigma=self.workingpars_sim[1]
+#        rho=self.workingpars_sim[2]
+        alpha=self.workingpars_sim[3] 
+        xi=self.workingpars_sim[4]
 
         theta=sigma*sigma
         eta=xi*sigma
@@ -629,77 +690,106 @@ class Heston_treeX2(svol_model):
         ret['rep_rho']=rho
         ret['rep_alpha']=alpha
         ret['rep_eta']=eta
-        ret['misc_q']=q
 #        ret['u0']=u0
-#        ret['uT']=uT
         ret['rep_v0']=v0
-        ret['rep_vT']=vT
+
+
+        sim_mu=self.workingpars_sim[0] 
+        sim_sigma=self.workingpars_sim[1]
+        sim_rho=self.workingpars_sim[2]
+        sim_alpha=self.workingpars_sim[3] 
+        sim_xi=self.workingpars_sim[4]
+        sim_u0=self.workingpars_sim[5]
+
+        sim_theta=sim_sigma*sim_sigma
+        sim_eta=sim_xi*sim_sigma
+        sim_v0=sim_sigma*sim_sigma*sim_u0
+
+        ret['sim_wrk_mu']=sim_mu
+        ret['sim_wrk_sigma']=sim_sigma
+        ret['sim_wrk_rho']=sim_rho
+        ret['sim_wrk_alpha']=sim_alpha
+        ret['sim_wrk_xi']=sim_xi
+        ret['sim_wrk_u0']=sim_u0
+
+        ret['sim_rep_mu']=sim_mu
+        ret['sim_rep_theta']=sim_theta
+        ret['sim_rep_rho']=sim_rho
+        ret['sim_rep_alpha']=sim_alpha
+        ret['sim_rep_eta']=sim_eta
+        ret['sim_rep_v0']=sim_v0
+
+
+        ret['misc_q']=q
+#        ret['uT']=uT
+        ret['misc_vT']=vT
         ret['ts_vpath']=vpath
         ret['ts_upath']=self.upath
+
+        Nret=self.Nobs-1
+        self.standardreturns=np.zeros(Nret)
+        Heston_standardreturn(self.yasset,self.upath,self.dt,rho,sigma,mu,alpha,xi,self.standardreturns)
+        ret['ts_standardreturns']=self.standardreturns
         
         return ret
 
-    def workingpars_update(self,workingpars):
-        super().workingpars_update(workingpars)
+    def update(self):
+        super().update()
         
-        if( self.current==False):
-# update all grid quantities to be cached:
-#TODO: struct-ify?
-            mu=self.workingpars[0] 
-            sigma=self.workingpars[1]
-            rho=self.workingpars[2]
-            alpha=self.workingpars[3] 
-            xi=self.workingpars[4]
-            u0=self.workingpars[5]
-            q=2.0*alpha/(xi*xi)
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+        u0=self.workingpars[5]
+        q=2.0*alpha/(xi*xi)
 
 #            dt=self.dt
-            dt2=self.dt2
-            
-            if( alpha*dt2 < 1.0e-6 ):
-                self.gamm=0.5*alpha*dt2*(1+0.5*alpha*dt2)
-                self.Vb=(xi*xi)*np.exp(-alpha*dt2)*dt2*(1-0.5*alpha*dt2)
-            else:
-                self.gamm=np.expm1(alpha*dt2)/2.0
-                self.Vb=-xi*xi*np.expm1(-alpha*dt2)*np.exp(-alpha*dt2)/alpha
-            if( np.isnan(self.gamm) or np.isnan(self.Vb) ):
-                print('NAN ERROR: gamm,Vb')
-                print(alpha,q,xi,self.gamm,self.Vb)
-            
-            self.delta=self.Vb/(2.0*self.lamb)
-            
-            t=np.linspace(1,2*self.Nobs,2*self.Nobs)*dt2
-            uj=-np.expm1(-alpha*t)+np.exp(-alpha*t)*u0
+        dt2=self.dt2
         
-            deltajrat=(np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-0.5)
-            self.nj=np.array( np.ceil( np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-np.sqrt(0.25+2.0*self.gamm/self.delta ) ) , int)
+        if( alpha*dt2 < 1.0e-6 ):
+            self.gamm=0.5*alpha*dt2*(1+0.5*alpha*dt2)
+            self.Vb=(xi*xi)*np.exp(-alpha*dt2)*dt2*(1-0.5*alpha*dt2)
+        else:
+            self.gamm=np.expm1(alpha*dt2)/2.0
+            self.Vb=-xi*xi*np.expm1(-alpha*dt2)*np.exp(-alpha*dt2)/alpha
+        if( np.isnan(self.gamm) or np.isnan(self.Vb) ):
+            print('NAN ERROR: gamm,Vb')
+            print(alpha,q,xi,self.gamm,self.Vb)
         
-            self.tu0=uj+self.delta*(-self.nj*deltajrat+self.nj*(self.nj-1)/2.0)
-            self.tdelta0=self.delta*(deltajrat-self.nj)
-            
-            maxindex = 2*self.Nobs+np.max(self.nj)
+        self.delta=self.Vb/(2.0*self.gridfactor)
+        
+        t=np.linspace(1,2*self.Nobs,2*self.Nobs)*dt2
+        uj=-np.expm1(-alpha*t)+np.exp(-alpha*t)*u0
+    
+        deltajrat=(np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-0.5)
+        self.nj=np.array( np.ceil( np.sqrt(0.25+2.0*(self.gamm+uj)/self.delta)-np.sqrt(0.25+2.0*self.gamm/self.delta ) ) , int)
+    
+        self.tu0=uj+self.delta*(-self.nj*deltajrat+self.nj*(self.nj-1)/2.0)
+        self.tdelta0=self.delta*(deltajrat-self.nj)
+        
+        maxindex = 2*self.Nobs+np.max(self.nj)
 
 # maxindex can depend on parameters, so have to reallocate all the time...
-            self.ni_s=np.zeros((2,maxindex+1),dtype=int)
-            self.space=np.zeros((7,maxindex+1))
-            self.pi_s=np.zeros((27,maxindex+1))
-            self.u_s=np.zeros((3,maxindex+1))
-            self.p_s=np.zeros((2,maxindex+1))
-            
-            CondProbCalc='Limit'
-            if( CondProbCalc=='Limit' ):
-                coeff_dt=(mu-alpha*rho*sigma/xi)*dt2
-                coeff_iu = rho*alpha*sigma/xi -sigma*sigma/2.0
-                coeff_du = rho*sigma/xi
-                self.condprob_calc=lambda yasset,u_prev,u_mid,u_this: condassetprob_calc_limit_X2(yasset,u_prev,u_mid,u_this,dt2,rho,sigma,coeff_dt,coeff_iu,coeff_du)
-            elif( CondProbCalc=='Exact' ):
-        #TODO: If you really want to use this then it needs to be cleaned up and reviewed!!
+        self.ni_s=np.zeros((2,maxindex+1),dtype=int)
+        self.space=np.zeros((7,maxindex+1))
+        self.pi_s=np.zeros((27,maxindex+1))
+        self.u_s=np.zeros((3,maxindex+1))
+        self.p_s=np.zeros((2,maxindex+1))
         
-                pass
-            else:
-                print('Unknown Calc -- Handle')                
+        CondProbCalc='Limit'
+        if( CondProbCalc=='Limit' ):
+            coeff_dt=(mu-alpha*rho*sigma/xi)*dt2
+            coeff_iu = rho*alpha*sigma/xi -sigma*sigma/2.0
+            coeff_du = rho*sigma/xi
+            self.condprob_calc=lambda yasset,u_prev,u_mid,u_this: condassetprob_calc_limit_X2(yasset,u_prev,u_mid,u_this,dt2,rho,sigma,coeff_dt,coeff_iu,coeff_du)
+        elif( CondProbCalc=='Exact' ):
+    #TODO: If you really want to use this then it needs to be cleaned up and reviewed!!
+    
+            pass
+        else:
+            print('Unknown Calc -- Handle')                
             
-
         return
 
     def calculate(self,OptRun=True):
@@ -810,17 +900,17 @@ class Heston_treeX2(svol_model):
                 p_this = p_s[0,0:n2_top]
                 iseven=True
     
-            condprob_up_mid[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min+1],u_this[ni2[ni1-n1_min+1]-n2_min])
-            condprob_up_dn[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min+1],u_this[ni2[ni1-n1_min+1]-n2_min-1])
-            condprob_up_up[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min+1],u_this[ni2[ni1-n1_min+1]-n2_min+1])
+            condprob_up_mid[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min+1],u_this[ni2[ni1-n1_min+1]-n2_min])
+            condprob_up_dn[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min+1],u_this[ni2[ni1-n1_min+1]-n2_min-1])
+            condprob_up_up[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min+1],u_this[ni2[ni1-n1_min+1]-n2_min+1])
     
-            condprob_mid_mid[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min],u_this[ni2[ni1-n1_min]-n2_min])
-            condprob_mid_dn[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min],u_this[ni2[ni1-n1_min]-n2_min-1])
-            condprob_mid_up[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min],u_this[ni2[ni1-n1_min]-n2_min+1])
+            condprob_mid_mid[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min],u_this[ni2[ni1-n1_min]-n2_min])
+            condprob_mid_dn[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min],u_this[ni2[ni1-n1_min]-n2_min-1])
+            condprob_mid_up[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min],u_this[ni2[ni1-n1_min]-n2_min+1])
     
-            condprob_dn_mid[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min-1],u_this[ni2[ni1-n1_min-1]-n2_min])
-            condprob_dn_dn[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min-1],u_this[ni2[ni1-n1_min-1]-n2_min-1])
-            condprob_dn_up[:]=self.condprob_calc(yasset[cc],u_prev,u_mid[ni1-n1_min-1],u_this[ni2[ni1-n1_min-1]-n2_min+1])
+            condprob_dn_mid[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min-1],u_this[ni2[ni1-n1_min-1]-n2_min])
+            condprob_dn_dn[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min-1],u_this[ni2[ni1-n1_min-1]-n2_min-1])
+            condprob_dn_up[:]=self.condprob_calc(yasset[[cc]],u_prev,u_mid[ni1-n1_min-1],u_this[ni2[ni1-n1_min-1]-n2_min+1])
     
             p_this[:]=np.zeros(n2_top)
             if( OptRun==False ):
@@ -971,6 +1061,14 @@ class Heston_treeX2(svol_model):
         self.calculate(False)
         return 
 
+    def status(self):
+        if self.current != True:
+            return 'Failure','Incorrect Model State.'
+
+        status = 'Success'
+        message = 'No issues.'        
+        return (status,message)
+
 #---------------------------------------------
 
 class Heston_grid(svol_model):
@@ -987,16 +1085,21 @@ class Heston_grid(svol_model):
         rho=0.0
         alpha=2.0
         xi=1.0
+        u0=1.0
+        
+        self.gridfactor=0.4
 
-        self.lamb=0.4
-
-        self.workingpars_names=['mu','sigma','rho','alpha','xi']
-        self.workingpars=np.array([mu,sigma,rho,alpha,xi])
-        self.workingpars_diffs=[0.0001,0.0001,0.0005,0.001,0.0001]
+        self.workingpars_names=['mu','sigma','rho','alpha','xi','u0']
+        self.workingpars=np.array([mu,sigma,rho,alpha,xi,u0])
+        self.workingpars_sim=np.array([mu,sigma,rho,alpha,xi,u0])
+        self.workingpars_diffs=[0.0001,0.0001,0.0005,0.001,0.0001,0.0001]
 
         alpha_min=np.minimum(2.0/((self.Nobs-1)/252.0),1.0)
 #                 [hmu, hsigma, rho, alpha, xi,u0]
-        self.workingpars_bounds=[(-0.5,0.5), (0.05, 0.5), (-0.9,0.9), (alpha_min, 20.0), (0.1, 4.0)]
+        self.workingpars_bounds=[(-0.5,0.5), (0.05, 0.5), (-0.9,0.9), (alpha_min, 20.0), (0.1, 4.0),(0.1,3.0)]
+
+        self.workingpars_optflag=[True for x in self.workingpars]
+        self.workingpars_optflag[-1]=False
 
         if 'init' in self.options:
             self.initpars_reporting(self.options['init'])
@@ -1017,10 +1120,12 @@ class Heston_grid(svol_model):
         rho=self.workingpars[2]
         alpha=self.workingpars[3] 
         xi=self.workingpars[4]
+        u0=self.workingpars[5]
 
         theta=sigma*sigma
         eta=xi*sigma
-
+        v0=u0*theta
+        
         for x in pardict:
             if( x=='mu' ):
                 mu=pardict[x]
@@ -1038,39 +1143,50 @@ class Heston_grid(svol_model):
 
         sigma=np.sqrt(theta)
         xi=eta/sigma
+        u0=v0/theta
 
-        self.workingpars[0]=mu
-        self.workingpars[1]=sigma
-        self.workingpars[2]=rho
-        self.workingpars[3]=alpha
-        self.workingpars[4]=xi
-        self.frozen_u0=v0/theta
-
+        if( pardict['type']=='init' ):
+            self.workingpars[0]=mu
+            self.workingpars[1]=sigma
+            self.workingpars[2]=rho
+            self.workingpars[3]=alpha
+            self.workingpars[4]=xi
+            self.workingpars[5]=u0
+        else:
+            self.workingpars_sim[0]=mu
+            self.workingpars_sim[1]=sigma
+            self.workingpars_sim[2]=rho
+            self.workingpars_sim[3]=alpha
+            self.workingpars_sim[4]=xi
+            self.workingpars_sim[5]=u0
+            
         return
 
     def get_structure(self):
         assetname='asset'
         variancename='variance'
 
-        sigma=self.workingpars[1]
-        rho=self.workingpars[2]
+        sigma=self.workingpars_sim[1]
+        rho=self.workingpars_sim[2]
+        u0=self.workingpars[5]
 
         corrmatrix=np.array([[1.0,rho],[rho,1.0]])
 #TODO: best choice based on pars?
         Nperstep=4
 
         assetval=1.0
-        varianceval=self.frozen_u0*sigma*sigma
+        varianceval=u0*sigma*sigma
 
         return assetname,assetval,variancename,varianceval,corrmatrix,Nperstep
 
     def sim_step(self,asset,variance,Zs):
 
-        mu=self.workingpars[0] 
-        sigma=self.workingpars[1]
-        rho=self.workingpars[2]
-        alpha=self.workingpars[3] 
-        xi=self.workingpars[4]
+        mu=self.workingpars_sim[0] 
+        sigma=self.workingpars_sim[1]
+#        rho=self.workingpars_sim[2]
+        alpha=self.workingpars_sim[3] 
+        xi=self.workingpars_sim[4]
+#        u0=self.workingpars[5]
 
         theta=sigma*sigma
         eta=xi*sigma
@@ -1100,6 +1216,7 @@ class Heston_grid(svol_model):
         rho=self.workingpars[2]
         alpha=self.workingpars[3] 
         xi=self.workingpars[4]
+        u0=self.workingpars[5]
 
         theta=sigma*sigma
         eta=xi*sigma
@@ -1121,56 +1238,84 @@ class Heston_grid(svol_model):
         ret['rep_rho']=rho
         ret['rep_alpha']=alpha
         ret['rep_eta']=eta
-        ret['misc_q']=q
 #        ret['u0']=u0
-#        ret['uT']=uT
         ret['rep_v0']=v0
-        ret['rep_vT']=vT
+
+        sim_mu=self.workingpars_sim[0] 
+        sim_sigma=self.workingpars_sim[1]
+        sim_rho=self.workingpars_sim[2]
+        sim_alpha=self.workingpars_sim[3] 
+        sim_xi=self.workingpars_sim[4]
+        sim_u0=self.workingpars_sim[5]
+
+        sim_theta=sim_sigma*sim_sigma
+        sim_eta=sim_xi*sim_sigma
+        sim_v0=sim_sigma*sim_sigma*sim_u0
+
+        ret['sim_wrk_mu']=sim_mu
+        ret['sim_wrk_sigma']=sim_sigma
+        ret['sim_wrk_rho']=sim_rho
+        ret['sim_wrk_alpha']=sim_alpha
+        ret['sim_wrk_xi']=sim_xi
+        ret['sim_wrk_u0']=sim_u0
+
+        ret['sim_rep_mu']=sim_mu
+        ret['sim_rep_theta']=sim_theta
+        ret['sim_rep_rho']=sim_rho
+        ret['sim_rep_alpha']=sim_alpha
+        ret['sim_rep_eta']=sim_eta
+        ret['sim_rep_v0']=sim_v0
+
+        
+        Nusefulgrid=Heston_UsefulGrid(0.001,self.grid_lnp_init)
+        ret['misc_Nusefulgrid']=Nusefulgrid
+
+        ret['misc_q']=q
+#        ret['uT']=uT
+        ret['misc_vT']=vT
         ret['ts_vpath']=vpath
         ret['ts_upath']=self.upath
+
+        Nret=self.Nobs-1
+        self.standardreturns=np.zeros(Nret)
+        Heston_standardreturn(self.yasset,self.upath,self.dt,rho,sigma,mu,alpha,xi,self.standardreturns)
+        ret['ts_standardreturns']=self.standardreturns
         
         return ret
     
-    def workingpars_update(self,workingpars):
-        super().workingpars_update(workingpars)
+    def update(self):
+        super().update()
 
-        if( self.current==False):
-# update all grid quantities to be cached:
-#TODO: struct-ify?
-            mu=self.workingpars[0] 
-            sigma=self.workingpars[1]
-            rho=self.workingpars[2]
-            alpha=self.workingpars[3] 
-            xi=self.workingpars[4]
-    
-            tmp=Heston_griddefs(self.dt,alpha,xi,self.lamb)
-            self.grid_i_lower=tmp[0]
-            self.grid_i_upper=tmp[1]
-            self.grid_i_map=tmp[2]
-            self.grid_u_grid=tmp[3]
-            self.grid_lnp_init=tmp[4]
-            self.grid_pi3=tmp[5]
-    
-            self.grid_i_length=self.grid_i_upper-self.grid_i_lower + 1
-    
-            self.grid_lnp_prev = np.zeros(self.grid_i_length)
-            self.grid_lnp_this = np.zeros(self.grid_i_length)
-    
-            Nret=self.Nobs-1
-    
-            self.grid_lncondprob_dn = np.zeros((Nret,self.grid_i_length))
-            self.grid_lncondprob_mid = np.zeros((Nret,self.grid_i_length))
-            self.grid_lncondprob_up = np.zeros((Nret,self.grid_i_length))
-    
-            self.grid_tmp_dn = np.zeros(self.grid_i_length)
-            self.grid_tmp_mid = np.zeros(self.grid_i_length)
-            self.grid_tmp_up = np.zeros(self.grid_i_length)
-    
-            lncondprob_calc=lambda yasset,u_prev,u_this,lncp: Heston_lncondassetprob(yasset,u_prev,u_this,self.dt,rho,sigma,mu,alpha,xi,lncp)
-    
-            lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map+1],self.grid_lncondprob_up)
-            lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map],self.grid_lncondprob_mid)
-            lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map-1],self.grid_lncondprob_dn)
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+
+        tmp=Heston_griddefs(self.dt,alpha,xi,self.gridfactor)
+        self.grid_i_lower=tmp[0]
+        self.grid_i_upper=tmp[1]
+        self.grid_i_map=tmp[2]
+        self.grid_u_grid=tmp[3]
+        self.grid_lnp_init=tmp[4]
+        self.grid_pi3=tmp[5]
+
+        self.grid_i_length=self.grid_i_upper-self.grid_i_lower + 1
+
+        self.grid_lnp_prev = np.zeros(self.grid_i_length)
+        self.grid_lnp_this = np.zeros(self.grid_i_length)
+
+        Nret=self.Nobs-1
+
+        self.grid_lncondprob_dn = np.zeros((Nret,self.grid_i_length))
+        self.grid_lncondprob_mid = np.zeros((Nret,self.grid_i_length))
+        self.grid_lncondprob_up = np.zeros((Nret,self.grid_i_length))
+
+        self.grid_tmp_dn = np.zeros(self.grid_i_length)
+        self.grid_tmp_mid = np.zeros(self.grid_i_length)
+        self.grid_tmp_up = np.zeros(self.grid_i_length)
+
+        self.lncondprob_calc=lambda yasset,u_prev,u_this,lncp: Heston_lncondassetprob(yasset,u_prev,u_this,self.dt,rho,sigma,mu,alpha,xi,lncp)
             
         return
 
@@ -1184,6 +1329,11 @@ class Heston_grid(svol_model):
         alpha=self.workingpars[3] 
         xi=self.workingpars[4]
 
+# needs to be here to avoid double calculate by derive classdes:
+        self.lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map+1],self.grid_lncondprob_up)
+        self.lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map],self.grid_lncondprob_mid)
+        self.lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map-1],self.grid_lncondprob_dn)
+    
         lnp_init=self.grid_lnp_init
         pi3=self.grid_pi3
 
@@ -1197,7 +1347,7 @@ class Heston_grid(svol_model):
         lncondprob_dn = self.grid_lncondprob_dn
         lncondprob_mid = self.grid_lncondprob_mid
         lncondprob_up = self.grid_lncondprob_up
-    
+
         tmp_dn = self.grid_tmp_dn
         tmp_mid = self.grid_tmp_mid
         tmp_up = self.grid_tmp_up
@@ -1230,6 +1380,12 @@ class Heston_grid(svol_model):
         alpha=self.workingpars[3] 
         xi=self.workingpars[4]
     
+# don'tneed to recalc on here?
+# needs to be here to avoid double calculate by derived classdes:
+        self.lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map+1],self.grid_lncondprob_up)
+        self.lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map],self.grid_lncondprob_mid)
+        self.lncondprob_calc(self.yasset,self.grid_u_grid,self.grid_u_grid[self.grid_i_map-1],self.grid_lncondprob_dn)
+
 #        i_lower=self.grid_i_lower
 #        i_upper=self.grid_i_upper
         i_length=self.grid_i_length
@@ -1310,15 +1466,19 @@ class Heston_grid(svol_model):
 
 # prediction of the variance at end of path:    
         upath=np.zeros(Nret+1)
+#        indpath=np.zeros(Nret+1,int)
         entry_current=np.argmax(lnp_this)
+#        indpath[Nret]=entry_current
         uT=u_grid[entry_current]
         upath[Nret]=uT
         for cpath in range (Nret-1,-1,-1):        
 #            print(cpath)
             entry_current=vilist[cpath][entry_current]
+#            indpath[cpath]=entry_current
             upath[cpath]=u_grid[entry_current]
 #        print(uT)
     
+
 # include in numfunevals even though this one will be a lot slower.
         self.numfunevals+=1
 # don't change current -- it should be before here, and if this 
@@ -1327,3 +1487,11 @@ class Heston_grid(svol_model):
         self.upath=upath
 
         return 
+
+    def status(self):
+        if self.current != True:
+            return 'Failure','Incorrect Model State.'
+
+        status = 'Success'
+        message = 'No issues.'        
+        return (status,message)
